# Copyright (C) 2020 Ludvig Larsson aka Valmond ludviglarsson@protonmail.com
# This code is licensed under the terms of the GNU Public License, version 3 or later (GPL-3.0-or-later)

import os
import sys
import time
import json

from struct import pack, unpack
from Cryptodome.PublicKey import RSA
import random

from socket_communication import SecureCommunication
import database_query as dq
import rsa_helper as rh
import database_connection as dc
from configuration import get_configuration_value
import address_registry as ar
import locator


# todo: randomly call requre_update_data_information for fingers


"""
The scheduler makes certain commands at certain intervals like:

* Trying to find new nodes
* Check if nodes are up (and update the DB accordingly)
* Remove stale nodes
* Check that all our data have 10 fingers, or try to get a new finger
* Verify fingers pointing on our data (and update the DB accordingly)
* Remove stale fingers

"""

"""
For convenience, we'll make the different functions in this file
Later on when there are several, we should move them to separate source files.

"""

_log_event_calls = False  # prints to stdout all the calls the scheduler does

###################################
#       DB helpers
#


# Updates the nodes quality
def update_node_quality(node_id, success):
    # Updates db::node: CONNECTION_TRIES, CONNECTION_FAILS, LAST_CONNECTION_UNIXTIME
    # get a random node:
    node_id = dq.get_random_node_id(only_validated_nodes=False)
    if node_id is not None:
        info = dq.get_node_data(node_id)
        sc = SecureCommunication(info[0], info[1], info[2]) #  IT IS NOT str(...) but ENCODE/DECODE!!!!!
        dq.update_node_quality(node_id, sc.is_active())
        sc.close()


###################################
#       NODES
#


def get_node_connection(node_id):
    tmp = dq.get_node_data(node_id)
    ip = str(tmp[0])
    port = tmp[1]
    rsa_pub_key = str(tmp[2])
    # print('Try to connect to random node:', ip, port) #, rsa_pub_key)
    # print('using rsa type, len', type(rsa_pub_key), len(rsa_pub_key))
    # print('using rsa', rsa_pub_key[-10:])
    sc = SecureCommunication(ip, port, RSA.importKey(bytes.fromhex(rsa_pub_key)))  # rsa_pub here is the one linked to the other Node
    return sc  # You check it's active!


# Returns a secure, functionnal node, or None
def get_random_node_connection(only_validated_nodes=True):
    node_id = dq.get_random_node_id(only_validated_nodes)  # todo: remove the False in the call
    if node_id is None:
        return None, 0
    # sc = SecureCommunication(ip, port, RSA.importKey(bytes.fromhex(rsa_pub_key)))  # rsa_pub here is the one linked to the other Node
    sc = get_node_connection(node_id)
    if sc.is_active():
        return sc, node_id
    return None, node_id


# One of the scheduled events:
def find_new_nodes():
    # Get a random nod in our db:node, connect to it and ask for some nodes
    sc, node_id = get_random_node_connection(False)  # todo: remove the False in the call

    if sc is None:
        # print('Random node is not up (Find new Nodes).')
        return
    # print('Ask for nodes from node', node_id)
    # Query for some random node(s)
    ip = get_configuration_value('ip')
    port = get_configuration_value('port')
    _, rsa_pub = rh.get_rsa_keypair(True)

    # Send which port we prefer
    port = str(port).encode()
    # and our public key
    rsa_pub = rsa_pub.exportKey('DER').hex().encode()

    # print('Send SEND_NODES req.', flush=True)
    json_answer = sc.send(b'SEND_NODES' + str(ip).encode() + b',' + port + b',' + rsa_pub)  # ,' + aa + b',' + bb + b',' + rr)  # expected return: json with a node: ip, port, rsa_pub_key, or an empty json if no nodes: {}
    # print('Send SEND_NODES req. Got Answer:', json_answer[:48], flush=True)
    if json_answer:
        try:
            # print('Got an answer for our SEND NODES, lets try to json decode it:')
            j = json.loads(json_answer.decode())  # decode to string and then put JSON-data to a variable
            # insert it in the database
            if 'ip' in j and 'port' in j and 'rsa_pub_key' in j:
                # convert ip & rsa_pub_key from bytes to str, port is an integer so no need.
                # print('Got node:', str(j['ip']), ':', j['port'])
                dq.insert_node(str(j['ip']), j['port'], str(j['rsa_pub_key']), 'From Scheduler SEND_NODES')
            else:
                print('#No information in json, bailing out:', json_answer.decode())
        except:
            print('JSON Exception!')
            print('tried to decode:', json_answer.decode())
            return
    else:
        print('Got empty answer. Bail out.')


# don't forget to use update_node_quality()


# One of the scheduled events, checks one node and
# updates its quality based on if we cqn connect to it or not
def check_node_quality():
    # update connection_tries & connection_fails
    # update last_connection_datetime if the connection worked
    sc, node_id = get_random_node_connection(False)  # todo: remove the False in the call
    node_up = False
    if sc is not None:
        node_up = sc.is_active()

    # print('### UPDATE NODE QUALITY: ', node_id, node_up)
    dq.update_node_quality(node_id, node_up)

    if node_up:
        sc.close()


# One of the scheduled events:
def remove_stale_nodes():
    # When we have a large number of nodes, cull the ones no longer working
    # by removing the oldest db:nodes::last_connection_datetime
    pass


def verify_our_address_in_nodes():
    # Verify all nodes know our current address
    our_ip = get_configuration_value('ip')
    our_port = get_configuration_value('port')

    # Find nodes who think our address is an old one
    node_id = dc.query_one("SELECT id FROM node WHERE our_ip != ? OR our_port != ? ORDER BY RANDOM() LIMIT 1", (our_ip, our_port))
    if node_id is not None:
        # Send UPDATEADDRESS requeest to node:
        sc = get_node_connection(node_id)
        if sc.is_active():

            their_our_ip = dc.query_one("SELECT our_ip FROM node WHERE id = ? ", (node_id, ))
            their_our_port = dc.query_one("SELECT our_port FROM node WHERE id = ? ", (node_id, ))

            print("STALE_ADDRESS: Found a node who has an old version of our address: id:", node_id, flush=True)
            print("STALE_ADDRESS: their address keps:", their_our_ip, their_our_port, flush=True)
            prv, pub = dq.get_rsa_hex_keys()
            _, rsa_pub_hex = dq.get_rsa_hex_keys()
            # Send our address update:
            print("STALE_ADDRESS: send our address:", our_ip, our_port, flush=True)
            digest_query = sc.send(b'NEWADDRESS' + str(our_ip).encode() + b',' + str(our_port).encode() + b',' + rsa_pub_hex.encode())

            if digest_query is not None:
                if sc.is_active():
                    digest = rh.create_digest(digest_query, RSA.importKey(bytes.fromhex(prv)))
                    # print('Okay all systems go!', flush=True)
                    #   <<        Send Recv           >>
                    answer = sc.send(digest)

                    if answer == b'OK':
                        # dc.update("UPDATE NODE SET our_ip = ?, out_port = ? WHERE id = ? LIMIT 1", (our_ip, our_port, node_id))
                        print("STALE_ADDRESS: update our db:  nodeid, our address", node_id, our_ip, our_port, flush=True)

                        dq.update_our_address_for_node(node_id, our_ip, our_port)
                        print("Our new address was successfully sent to node " + str(node_id) + " and then updated locally.")
                    else:
                        print("Our new address sent to node " + str(node_id) + " was REJECTED because of wrong digest")
                else:
                    print("Our new address sent to node " + str(node_id) + " was REJECTED because node closed connection unexpectedly")
                    if digest_query is not None:
                        print("however they sent back this:", digest_query)
            else:
                print("Our new address sent to node " + str(node_id) + " was REJECTED because node did not even answer our request call")

#

###################################
#       Our Data
#


# Call this when user has updated its data (so it is a newer version)
# We will ask the finger(s node) to accept a newer version of the data shared
def _update_finger_data(finger_id, node_id, data_id, new_data_version, file_size):
    # print('Try to update finger id:', finger_id)
    tmp = dq.get_node_data(node_id)
    ip = str(tmp[0])
    port = tmp[1]
    other_rsa_pub_key = str(tmp[2])
    sc = SecureCommunication(ip, port, RSA.importKey(bytes.fromhex(other_rsa_pub_key)))
    if not sc.is_active():
        # node is not up
        #print('Could not connect for update to ', ip, ":", port, ' so bail out.')
        return False

    # get data information & data
    my_name, my_original_file_name = dq.get_link_data_by_dataid(data_id)

    # get our PubKey:
    prv, pub = rh.get_rsa_keypair()
    _, rsa_pub_hex = dq.get_rsa_hex_keys()
    # rsa_pub_hex = pub.exportKey('DER').hex().encode()

    nonce = dq.get_nonce_from_dataid(data_id)
    # ask for an update
    query = b'UPDATEDATA' + pack('>Q', file_size) + b',' + my_name.encode() + b',' +\
            rsa_pub_hex.encode() + b',' + str(new_data_version).encode() + b',' + str(nonce).encode() + b','

    random_checksum = sc.send(query)
    if not random_checksum:
        print('Got no answer')
        sc.close()
        return False

    # make a digest and send back for authentification, prove we actually own the data to be updated:
    digest = rh.create_digest(random_checksum, prv)
    # print('Calculating digest:', digest)
    # send back digest
    answer = sc.send(digest)
    if not answer:
        print('No answer for digest request, bail out.')
        return False
    if answer[:10] != b'ACCEPTING!':

        if answer[:10] == b'NO_NEED_TO':  # They are up to date
            print("They already store the latest version, update our db.")
            dc.update("UPDATE finger SET data_version = ? WHERE id = ?", (new_data_version, finger_id))
            return True
        else:
            print('Refresh data query was turned down for our [' + my_name + ']:', answer[:10], ' node_id=', node_id, ' finger_id=', finger_id)
            # As they visibly no longer share our data lets remove the finger:
            dq.delete_finger(finger_id)
            sc.close()
            return False
    # Refresh accepted, lets go through with it:

    # ip = get_configuration_value('ip')
    # get our PubKey:
    if not prv:
        print('No prv.')
        return False

    with open('data/local/' + my_name + '.aes', 'rb') as f:
        file_data = f.read()
    if file_data is None:
        print('Refresh: File not found', my_original_file_name)
        sc.close()
        return False

    # other_ip, other_port, other_pub_key = dq.get_node_data(node_id)
    answer = sc.send(file_data)
    if answer != b'OK':
        #print('Refresh sending data failed')
        return False

    sc.close()
    #print('Refresh successful, do this sql:')
    # print("UPDATE finger SET data_version = ? WHERE id = ?", (new_data_version, finger_id))
    dc.update("UPDATE finger SET data_version = ? WHERE id = ?", (new_data_version, finger_id))
    #print('SQL done.')
    return True


# Try to get a new swicharoo for one of our data. Private function used by check_data_fingers()
# Returns true on success, otherwise False
def _get_new_finger(data_id):
    # As socket communication can fail in so many ways and we even know we'll be having stale addresses in our database, make a big try-catch here:
    sc = None
    try:
        # Connect to a random node
        # todo: only select nodes that do not store our data already!
        sc, node_id = get_random_node_connection(False)  # todo: remove the False in the call (use only verified nodes)
        if sc is None:
            return False

        # get our PubKey, ip, port
        prv, pub = dq.get_rsa_hex_keys()
        if not prv:
            sc.close()
            return False
        ip = get_configuration_value('ip')
        port = get_configuration_value('port')

        # get data information
        my_name, my_original_file_name = dq.get_link_data_by_dataid(data_id)
        my_version = dq.get_owned_data_version(data_id)

        # print('Check we do not already have the finger (node_id, data_id)', node_id, data_id)
        # Verify we are not already have a finger with this node:
        # todo: don't ask nodes that already store our data, this seems it isn't enough, maybe as sometimes we accidentally maskerade with different ip:ports
        # Achtually it seems like we check IP:PORT and not the actual RSA Pubkey?
        # Or just we contact the same node, but over another IP (like a 192. ... instead of public IP) we think they are different
        # but they are actually the same => they already store our data
        finger_id = dq.get_finger_id(node_id, data_id)
        # print('Scheduler: We got a finger_id:', finger_id)
        if finger_id is not None:  # This node already stores the data
            # print('Scheduler: We already have data stored at this node')
            sc.close()
            return False

        # print('Scheduler: Ask for new finger for data_id='+str(data_id)+' node_id='+str(node_id)+ ' data name='+my_name+' (origname='+my_original_file_name+')')
        # get the actual datasize from the data_id
        file_size = dq.get_owned_file_size(data_id)

        my_nonce = dq.get_nonce_from_dataid(data_id)

        # ask for a swicharoo
        query = b'SWICHAROO?' + pack('>Q', file_size) + pub.encode() + b'|' + my_name.encode() + b'|' + ip.encode() + b'|' +\
                str(port).encode() + b'|' + str(my_version).encode() + b'|' + str(my_nonce).encode()
        #   <<        Send Recv           >>
        answer = sc.send(query)
        if answer[:10] != b'YES_PLEASE':
            if answer[:10] == b'ALREADY_DO':
                # They already store our data, duh
                print('Scheduler: SWICHAROO query was turned down because they already store our data, node_id, data_id:', node_id, data_id)
                pass
            else:
                if answer[:10] == b'SRYNODATA.':
                    print('They turned down because they have nothing to share.')
                else:
                    print('Scheduler: SWICHAROO query was turned down, reason:', answer)
            sc.close()  # Close as this node does not behave well right now
            return False

        # Get size other node want us to store and check if we are OK:
        data = answer[10:]  # Remove the b'YES_PLEASE'
        data_size = data[:8]
        data = data[8:]  # Remove the size
        other_file_size = unpack('>Q',  data_size)[0]
        # print('other_file_size:', other_file_size, flush=True)

        if other_file_size > get_configuration_value('swicharoo_file_maxsize_kb')*1024:
            sc.send(b'U_TOO_BIG!')
            sc.close()  # Close as the node has no more space anyways
            print('Scheduler: their file is too big (max=' + str(get_configuration_value('swicharoo_file_maxsize_kb')*1024) + '): their file size=', other_file_size)
            return False

        split = data.split(b'|')
        if len(split) != 3:
            sc.send(b'ERROR_LEN!')
            print('Split should be 3, is:', len(split))
            sc.close()  # Close as the node has no more space anyways
            return False

        # Get all information from the other node
        other_name = split[0].decode()
        other_original_file_name = split[1].decode()
        # print('other_name:', other_name)
        digest_query = split[2]  # probably do not decode() this but use as is

        if digest_query is None or not sc.is_active():
            print('Other node refused our swicharoo query, bailing out too.', flush=True)
            sc.close()
            return False

        # If we should do it this way, we should also discern nodes by rsa,ip,port? or have several ip:port for 1 node...
        other_ip, other_port, other_pub_key = dq.get_node_data(node_id)  # Why do we both accept these data from the listener, but also get them from our database?

        # Check if we already store the data:
        if dq.get_link_data_by_name_and_rsapub(other_name, other_pub_key):  # We already store this data for this node!
            print('Scheduler: our swicharoo failed as we already store the data the other node proposes to switch.')
            sc.send(b'DATA_AGAIN')
            sc.close()
            return False

        # print('Do digest, digest_query=', digest_query, flush=True)
        # print('Do digest type=', type(digest_query), flush=True)

        digest = rh.create_digest(digest_query, RSA.importKey(bytes.fromhex(prv)))
        print('Scheduler: Okay all systems go, change our ' + my_name + ' for theirs: ' + other_name, flush=True)
        #   <<        Send Recv           >>
        other_data = sc.send(digest)

        # Save their data
        file_folder = 'data/' + dq.get_folder(other_pub_key) + '/'
        if not os.path.exists(file_folder):
            os.makedirs(file_folder)
        save_name = file_folder + other_name + '.aes'
        # print('Scheduler: Swicharoo, save off others file:', save_name, flush=True)
        with open(save_name, 'wb') as f:
            f.write(other_data)

        # I now have other: pubkey, name, origname, data, save off data and update DB
        dq.insert_data_shared(other_name, other_pub_key)  # Others data is now stored here for retrieval (by name + pubkey), deprecated other_original_file_name

        # print('Got their data, send ours:', flush=True)

        # send our data, expect b'ok' back
        # get the data
        with open('data/local/' + my_name + '.aes', 'rb') as f:
            file_data = f.read()
        if file_data is None:
            print('Scheduler: Our original file not found:', my_original_file_name, flush=True)
            sc.close()
            return False

        # Send our data
        #   <<        Send Recv           >>
        ack = sc.send(file_data)
        # print('Got final ACK:', ack, flush=True)

        # Insert new finger for my data
        dq.insert_finger(node_id, data_id)
        print("Scheduler: The Swicharoo request was a success; inserted a finger. NodeId, DataId:", node_id, data_id, flush=True)

        # As an address have now changed (the data has one more address),
        # updates tr & sub files according to what kind of data this is (base_data, TR, SUB)
        ar.refresh_data_addresses(data_id)

        sc.close()
        return True

    except:
        if sc is not None:
            sc.close()
    return False


# One of the scheduled events:
def check_data_fingers():  # Tries to get a new finger for data that does not have enough fingers
    # Get data that is not yet shared at all
    data_id = dq.get_random_not_shared_data()
    if data_id is not None:
        # Try to get 2 new fingers for a random data (if needed)
        check_data_finger(data_id, 2)

    # Just get a random node
    data_id = dq.get_random_local_data_id()
    if data_id is not None:
        # Try to get 2 new fingers for a random data (if needed)
        check_data_finger(data_id, 2)


# Tries 'tries' times to get a new finger for this data, if needed
def check_data_finger(data_id, tries=10, print_output=True):
    newly_acquired_fingers = 0
    fingers = dq.get_number_of_fingers(data_id)
    wanted_fingers = dq.get_wanted_fingers(data_id)
    if wanted_fingers <= 0:
        return
    while fingers < wanted_fingers and tries > 0:
        if _get_new_finger(data_id):
            fingers += 1
            newly_acquired_fingers += 1
            if print_output:
                print('New finger acquired (' + str(fingers) + '/' + str(wanted_fingers) + ')', flush=True)
        tries = tries - 1


# Get a finger and check if it carries our data
# If the finger is up
# Check version of data and update if needed
# If it does no longer carry our data, remove the finger
def _check_finger_carries_data(sc, my_name, data_rsa_pub, shared_data_id, finger_id):
    try:
        data_version = dq.get_owned_data_version(shared_data_id)
        filepath = os.path.join(locator.create_path('data/local/'), (my_name + '.aes'))
        with open(filepath, 'rb') as f:
            f.seek(0, os.SEEK_END)
            filesize = f.tell()

            if filesize > 0:
                print("file size", filesize, flush=True)

                s_verification_chunk = 256

                chunk_size = min(s_verification_chunk, filesize)
                # Get a random starting point
                start_byte = 0
                a = filesize - chunk_size
                if a > 0:
                    start_byte = random.randint(0, a)
                print("start_byte, file size, chunk_size", start_byte, filesize, chunk_size, flush=True)

                read_bytes = chunk_size

                query = 'ONEFINGER:' + my_name + ',' + data_rsa_pub + ',' + str(start_byte) + ',' + str(read_bytes)  # rsa in plaintext here as we use it as a name
                print("QUERY:", query, flush=True)
                their_version = sc.send(query.encode())

                success = False

                # They are supposed to send back their version
                if their_version:
                    their_version = their_version.decode()
                    if their_version.isdigit():
                        their_version = int(their_version)

                        if their_version >= data_version:
                            # Get the data
                            encrypted_data = sc.send(b'GO')
                            if encrypted_data is not None:
                                if len(encrypted_data) == read_bytes:
                                    print("Read our bytes and verify", flush=True)
                                    # Set back file pointer to start_byte, according to the beginning of the file
                                    f.seek(start_byte, os.SEEK_SET)
                                    verify = f.read(read_bytes)

                                if encrypted_data == verify:
                                    success = True

                            else:
                                pass  # They got a stale version, take a note / update
                                # todo: make a champ in the fingers DB "had_stale_version" and update it here

                if success:
                    print('Check Node Finger: Finger [' + str(finger_id) + '] is still sharing our data [' + my_name + ']!', flush=True)
                else:
                    print('Check Node Finger: Our data [' + my_name + '] is not (correctly) shared by the finger [' + str(finger_id) + '] ...', flush=True)

                # Update statistics
                dq.update_finger_data_integrity_quality(finger_id, success)


    except:
        print('Verification of data, some error for data:', filepath, flush=True)


# One of the scheduled events:
# Checks if the node containing this finger responds
# Close to check_node_quality except it runs for the lifetime
# of the finger, not of the node
def check_finger_quality():  # Take one random finger and check if it's responding, update DB accordingly
    # take a random finger and call up it's node
    # don't forget to use update_node_quality() AND update_finger_quality()
    finger_id = dq.get_random_finger_id()
    if finger_id is not None:
        # Get its node:
        # print('check_finger_quality', finger_id)
        node_id = dq.get_fingers_node_id(finger_id)
        # print('check_finger_quality node:', node_id)
        if node_id is not None:
            tmp = dq.get_node_data(node_id)
            if tmp:
                other_ip = str(tmp[0])
                other_port = tmp[1]
                other_rsa_pub_key = str(tmp[2])  # Was this the bug??! Check and keep only this if so
                #rsa_pub_key = RSA.importKey(bytes.fromhex(rsa_hex_pub_key))
                sc = SecureCommunication(other_ip, other_port, RSA.importKey(bytes.fromhex(other_rsa_pub_key)))  # rsa_pub here is the one linked to the other Node
                dq.update_finger_quality(node_id, finger_id, sc.is_active())

                if sc.is_active():
                    finger_data = dq.get_finger_data(finger_id)
                    data_id = finger_data[1]
                    my_name, my_origfilename = dq.get_link_data_by_dataid(data_id)
                    if my_name:
                        print("data_id, my_name, my_origfilename = ", data_id, my_name, my_origfilename)
                        _, my_rsa_pub_key = dq.get_rsa_hex_keys()
                        _check_finger_carries_data(sc, my_name, my_rsa_pub_key, data_id, finger_id)

                sc.close()
            else:
                print('Warning: node', node_id, ' has no information')
        else:
            print('Warning: finger', finger_id, 'points on non existing node:', node_id)


# One of the scheduled events:
def remove_stale_fingers():  # Sometimes we'll have to let go of a finger because it doesn't work well enough
    # For a random data that has 10 fingers
    # remove the oldest db:nodes::last_connection_datetime, if older than some day

    pass


# Here we check if user has updated data we own, and that other nodes share
# If so, lets try to send the new data to the nodes
def refresh_stale_fingers():
    stale_fingers_data = dq.get_stale_data_fingers()
    for finger_data in stale_fingers_data:
        finger_id = finger_data[0]
        node_id = finger_data[1]
        data_id = finger_data[2]
        data_version = finger_data[3]  # The fingers (old) version
        new_data_version = finger_data[4]

        #print('STALE FINGER id=', finger_id, ' node_id:', node_id, ' data_id:', data_id, 'data_version:', data_version, 'new_data_version:', new_data_version)
        # Ask for a refresh
        file_size = dq.get_owned_file_size(data_id)
        _update_finger_data(finger_id, node_id, data_id, new_data_version, file_size)


# Polled by scheduler
def check_address_translation_files():
    from insert_data import insert_file
    # NOTE: WE DO NOT check if the base datas addresses have changed
    # so we must update the tr & sub files when that happens
    # Reason: we don't know the old addresses
    # Solution: we must check when we get a new finger or when a fingers address changes and
    # update the corresponding tr & sub files

    # Get a _tr file
    address_translation_name = dc.query_one('SELECT name FROM data_owned WHERE address_translation_file == 1 ORDER BY RANDOM() LIMIT 1;')
    if address_translation_name:
        # todo: get the name better ffs
        name = address_translation_name[:len(address_translation_name) - len(ar.generate_trans_name(""))]
        # Check if our addresses have changed
        # With changed means addresses to nodes pointing to our payload

        file_name = 'trans/' + name + '_sub.atf'
        # print("Get addresses from file_name:", file_name, flush=True)
        # Compare the other address list in the _sub file with nodes pointing on this _tr file
        other_addresses_saved = ar.load_adresses(file_name, other_addresses=True)  # Saved off other-address list to download _tr file
        other_addresses_in_db = ar.get_all_addresses(ar.generate_trans_name(name))

        addresses_saved = ar.load_adresses(file_name, other_addresses=False)  # Saved off address list to download _tr file
        addresses_in_db = ar.get_all_addresses(name)

        if other_addresses_saved != other_addresses_in_db or addresses_saved != addresses_in_db:   # Shouldn't we also check the 'normal' addresses haven't changed ?!!
            # Update the _sub file
            print("Updating address substitution file for file: " + str(name), flush=True)
            # print("Old addresses (from file):", addresses_saved, flush=True)
            # print("New addresses (from DB  ):", addresses_in_db, flush=True)
            ar.set_addresses(name, use_tr_file=False)

            # The (substitute) file to reincrypt
            sub_name = ar.generate_sub_name(name)
            sub_file_name = ar.generate_sub_file_name(name)  # name => name_sub.atf

            insert_file(sub_name, ar.get_translation_folder(), sub_file_name)

            #update_data_main(['', sub_name, ar.get_translation_folder(), sub_file_name])

    # Get a _sub file
    address_substitution_name = dc.query_one('SELECT name FROM data_owned WHERE address_translation_file == 2 ORDER BY RANDOM() LIMIT 1;')
    if address_substitution_name:
        # todo: get the name better ffs
        name = address_substitution_name[:len(address_substitution_name) - len(ar.generate_sub_name(""))]
        # Check if our addresses have changed
        # With changed means addresses to nodes pointing to our payload

        file_name = 'trans/' + name + '_tr.atf'
        # print("Get addresses from file_name:", file_name, flush=True)
        # Compare the other address list in the _sub file with nodes pointing on this _tr file
        other_addresses_saved = ar.load_adresses(file_name, other_addresses=True)
        other_addresses_in_db = ar.get_all_addresses(ar.generate_sub_name(name))

        addresses_saved = ar.load_adresses(file_name, other_addresses=False)
        addresses_in_db = ar.get_all_addresses(name)

        if other_addresses_saved != other_addresses_in_db or addresses_saved != addresses_in_db:
            # Update the _tr file
            print("Updating address translation file for file: " + str(name), flush=True)
            # print("Old addresses (from file):", addresses_saved, flush=True)
            # print("New addresses (from DB  ):", addresses_in_db, flush=True)
            ar.set_addresses(name, use_tr_file=True)

            # The (translation) file to reincrypt
            trans_name = ar.generate_trans_name(name)
            trans_file_name = ar.generate_trans_file_name(name)  # name => name_tr.atf

            insert_file(trans_name, ar.get_substitution_folder(), trans_file_name)

            #update_data_main(['', trans_name, ar.get_translation_folder(), trans_file_name])


###################################
#       The actual Scheduling mechanism
#
def run_scheduler(run_continously=True):
    wait_time_msec = get_configuration_value('scheduler_wait_msec')
    wait_time = wait_time_msec * 0.001
    first_run = True
    while run_continously or first_run:
        first_run = False

        # Nodes
        if _log_event_calls:
            print('-------------------------------------------------------- 1')
            print('[Nodes]#Find new nodes:')
        find_new_nodes()

        if _log_event_calls:
            print('-------------------------------------------------------- 2')
            print('[Nodes]#Check if node is up and update node quality')
        check_node_quality()

        if _log_event_calls:
            print('-------------------------------------------------------- 3')
            print('[Nodes]#Prune stale nodes')
        remove_stale_nodes()

        if _log_event_calls:
            print('-------------------------------------------------------- 4')
            print('[Nodes]#Check if nodes know our address')
        verify_our_address_in_nodes()

        # Data
        if _log_event_calls:
            print('******************************************************** 5')
            print('[Data]#Check if data has enough fingers else try to get new ones')
        check_data_fingers()

        if _log_event_calls:
            print('******************************************************** 6')
            print('[Data]#Check if fingers node is up, updated and still carries our data')
        check_finger_quality()

        if _log_event_calls:
            print('******************************************************** 7')
            print('[Data]#Prune stale fingers')
        remove_stale_fingers()

        if _log_event_calls:
            print('******************************************************** 8')
            print('[Data]#Refresh stale owned data in nodes sharing older versions')
        refresh_stale_fingers()

        # todo: Figure out if we should do some polling approach to the translation files,
        # like check them out from time to time, or rely on that we got it right
        # updating address translation file when we insert a finger, change addres, node change address...

        # for now, poll address translation files
        if _log_event_calls:
            print('******************************************************** 9')
            print('[Data]#Refresh stale owned data in nodes sharing older versions')
        check_address_translation_files()

        # Don't use all CPU
        time.sleep(wait_time)



class Scheduler:
    def __init__(self):
        print("Starting scheduler:")
        self.continue_to_run = True
        self.wait_time = 1.0

    def forcestop(self):
        print("Shutting down scheduler...")
        self.continue_to_run = False

    def run(self):
        while self.continue_to_run:
            # We're in a thread here (hence the class) so just run once, then see if we should shut down or run again
            run_scheduler(run_continously=False)
            time.sleep(self.wait_time)


def test():
    for i in range(1, 9):
        dq.update_node_quality(1, True)
    for i in range(1, 5):
        dq.update_node_quality(1, False)


# This will only happen if you run the file, not if you import it
if __name__ == '__main__':
    if len(sys.argv) > 1 and sys.argv[1] == '--testrun':
        print('[scheduler.py] Testrun')
        test()
        sys.exit(0)

    if len(sys.argv) > 1 and sys.argv[1] == '--multitestrun':
        print('[scheduler.py] Multitestrun')
        _log_event_calls = True
        run_scheduler()
    else:
        run_scheduler()
